#include "byte_stream.h"

byte_stream::byte_stream(const std::string& buffer)
	:buffer_(buffer),
	read_position_(0),
	refcount_(1)
{
}

byte_stream* byte_stream::create(const std::string& buffer)
{
	return new byte_stream(buffer);
}

HRESULT STDMETHODCALLTYPE byte_stream::QueryInterface(REFIID iid, void ** ppvObject)
{
	if (iid == __uuidof(IUnknown)
		|| iid == __uuidof(IStream)
		|| iid == __uuidof(ISequentialStream))
	{
		*ppvObject = static_cast<IStream*>(this);
		AddRef();
		return S_OK;
	}
	else
	{
		return E_NOINTERFACE;
	}
}

ULONG STDMETHODCALLTYPE byte_stream::AddRef(void)
{
	return (ULONG)InterlockedIncrement(&refcount_);
}

ULONG STDMETHODCALLTYPE byte_stream::Release(void)
{
	ULONG res = (ULONG) InterlockedDecrement(&refcount_);
	if (res == 0)
		delete this;
	return res;
}

HRESULT STDMETHODCALLTYPE byte_stream::Read(void* pv, ULONG cb, ULONG* pcbRead)
{
	size_t bytes_left;
	size_t bytes_out;

	if(pcbRead != NULL)
		*pcbRead = 0;

	if(read_position_ == buffer_.size())
		return HRESULT_FROM_WIN32(ERROR_END_OF_MEDIA);

	bytes_left = buffer_.size() - read_position_;
	bytes_out = std::min<size_t>(cb, bytes_left);
	memcpy(pv, &buffer_[read_position_], bytes_out);
	read_position_ += bytes_out;
	if(pcbRead != NULL)
		*pcbRead = bytes_out;

	return S_OK;
}

HRESULT STDMETHODCALLTYPE byte_stream::Write(void const* pv, ULONG cb, ULONG* pcbWritten)
{
	size_t prev_size;

	if(cb <= 0)
	{
		if(pcbWritten != NULL)
			*pcbWritten = 0;

		return S_OK;
	}

	prev_size = buffer_.size();
	buffer_.resize(prev_size + cb);
	memcpy(&buffer_[prev_size], pv, cb);
	if(pcbWritten != NULL)
		*pcbWritten = cb;

	return S_OK;
}

HRESULT STDMETHODCALLTYPE byte_stream::SetSize(ULARGE_INTEGER)
{
	return E_NOTIMPL;
}

HRESULT STDMETHODCALLTYPE byte_stream::CopyTo(IStream*, ULARGE_INTEGER, ULARGE_INTEGER*, ULARGE_INTEGER*)
{
	return E_NOTIMPL;
}

HRESULT STDMETHODCALLTYPE byte_stream::Commit(DWORD)
{
	return E_NOTIMPL;
}

HRESULT STDMETHODCALLTYPE byte_stream::Revert(void)
{
	return E_NOTIMPL;
}

HRESULT STDMETHODCALLTYPE byte_stream::LockRegion(ULARGE_INTEGER, ULARGE_INTEGER, DWORD)
{
	return E_NOTIMPL;
}

HRESULT STDMETHODCALLTYPE byte_stream::UnlockRegion(ULARGE_INTEGER, ULARGE_INTEGER, DWORD)
{
	return E_NOTIMPL;
}

HRESULT STDMETHODCALLTYPE byte_stream::Clone(IStream **)
{
	return E_NOTIMPL;
}

HRESULT STDMETHODCALLTYPE byte_stream::Seek(LARGE_INTEGER liDistanceToMove, DWORD dwOrigin, ULARGE_INTEGER* lpNewFilePointer)
{
	size_t start = 0;
	int32_t new_read_position;

	switch(dwOrigin)
	{
	case STREAM_SEEK_SET:
		start = 0;
		break;
	case STREAM_SEEK_CUR:
		start = read_position_;
		break;
	case STREAM_SEEK_END:
		start = buffer_.size();
		break;
	default:
		return STG_E_INVALIDFUNCTION;
		break;
	}

	new_read_position = start + static_cast<int32_t>(liDistanceToMove.QuadPart);
	if (new_read_position < 0 || new_read_position > static_cast<int32_t>(buffer_.size()))
		return STG_E_SEEKERROR;

	read_position_ = static_cast<uint32_t>(new_read_position);

	if (lpNewFilePointer != NULL)
		lpNewFilePointer->QuadPart = read_position_;

	return S_OK;
}

HRESULT STDMETHODCALLTYPE byte_stream::Stat(STATSTG* pStatstg, DWORD grfStatFlag)
{
	if(pStatstg == NULL)
		return STG_E_INVALIDFUNCTION;

	ZeroMemory(pStatstg, sizeof(STATSTG));
	pStatstg->cbSize.QuadPart = buffer_.size();
	return S_OK;
}
